import unittest
import os
from adamspy.postprocess import xml
from adamspy.postprocess import launch_ppt
from adamspy.postprocess import _get_model_name_from_cmd
from adamspy.postprocess.ppt import get_lunar_results, SCRIPT_NAME

from test import *

class Test_ShrinkResults(unittest.TestCase):

    def setUp(self):
        # Delete existing shrunk results file
        if os.path.exists(TEST_EXISTING_RES_FILE.replace('.res', '') + xml.SHRUNK_RES_SUFFIX +'.res'):
            os.remove(TEST_EXISTING_RES_FILE.replace('.res', '') + xml.SHRUNK_RES_SUFFIX +'.res')
        
        # Create shrunk results file
        xml.shrink_results(TEST_EXISTING_RES_FILE, TEST_REQUESTS_TO_GET, TEST_REQUEST_TIME_MIN, TEST_REQUEST_TIME_MAX)
        
    def test_shrink_results_exist(self):
        """Tests if the shrunken results file exists
        
        """        
        # Check that the shrunk results file exists
        results_exist = os.path.exists(TEST_EXISTING_RES_FILE.replace('.res', '') + xml.SHRUNK_RES_SUFFIX +'.res')
        
        # Assert that the shrunk results file exists
        self.assertTrue(results_exist)

    def tearDown(self):
        # Delete shrunk results file
        if os.path.exists(TEST_EXISTING_RES_FILE.replace('.res', '') + xml.SHRUNK_RES_SUFFIX +'.res'):
            os.remove(TEST_EXISTING_RES_FILE.replace('.res', '') + xml.SHRUNK_RES_SUFFIX +'.res')

class Test_GetResults(unittest.TestCase):

    def setUp(self):
        # Get results from a results file
        self.requests, self.units = xml.get_results(TEST_EXISTING_RES_FILE, TEST_REQUESTS_TO_GET, TEST_REQUEST_TIME_MIN, TEST_REQUEST_TIME_MAX, return_units=True)

    def test_get_results_requests(self):       
        """Tests if `requests` returned by :meth:`xml.get_results` is as expected.

        """
        self.assertDictEqual(self.requests, TEST_EXPECTED_REQUESTS)

    def test_get_results_units(self):       
        """Tests if `units` returned by :meth:`xml.get_results` is as expected.

        """
        self.assertDictEqual(self.units, TEST_EXPECTED_REQUEST_UNITS)
    
    def test_get_results_all(self):    
        _requests, _units = xml.get_results(TEST_EXISTING_RES_FILE, t_min=TEST_REQUEST_TIME_MIN, t_max=TEST_REQUEST_TIME_MAX, return_units=True)
        
    def tearDown(self):
        return

class Test_GetResultsIterparse(unittest.TestCase):

    def setUp(self):
        # Get results from a results file
        self.requests, self.units = xml.get_results(TEST_EXISTING_RES_FILE, TEST_REQUESTS_TO_GET, TEST_REQUEST_TIME_MIN, TEST_REQUEST_TIME_MAX, return_units=True, use_iterparse=True)

    def test_get_results_requests(self):       
        """Tests if `requests` returned by :meth:`xml.get_results` is as expected.

        """
        self.assertDictEqual(self.requests, TEST_EXPECTED_REQUESTS)

    def test_get_results_units(self):       
        """Tests if `units` returned by :meth:`xml.get_results` is as expected.

        """
        self.assertDictEqual(self.units, TEST_EXPECTED_REQUEST_UNITS)
    
    def test_get_results_all(self):    
        _requests, _units = xml.get_results(TEST_EXISTING_RES_FILE, t_min=TEST_REQUEST_TIME_MIN, t_max=TEST_REQUEST_TIME_MAX, return_units=True, use_iterparse=True)
        
    def tearDown(self):
        return

class Test_LaunchPPT(unittest.TestCase):

    def setUp(self):
        return

    def test_get_model_name_from_cmd(self):
        mod_name = _get_model_name_from_cmd(TEST_EXISTING_CMD_FILE)
        self.assertEqual(mod_name, TEST_EXPECTED_MODNAME_IN_CMD)

    def test_pptAS_exists(self):
        directory = launch_ppt(TEST_EXISTING_RES_FILE)
        expected_directory = os.path.split(TEST_EXISTING_RES_FILE)[0]
        self.assertEqual(directory, expected_directory)

    def tearDown(self):
        return
        
class Test_GetResultsAview(unittest.TestCase):

    def setUp(self):
        self.res_files = [TEST_EXISTING_GENERIC_RES_FILE]
        self.output_file = os.path.join(os.path.dirname(TEST_EXISTING_GENERIC_RES_FILE), 'lunar_Y_file.csv')
        self.reqs_to_get = {'JOINT_1': ['FY']}
        self.t_min = 10
        self.t_max = 19.9


    def test_python_script_contents(self):
        """Checks that the python script that gets passed to aview is written correctly.

        """
        get_lunar_results(self.res_files, self.reqs_to_get, self.t_min, self.t_max, self.output_file, _just_write_script=True)
        directory = os.path.dirname(self.res_files[0])
        
        failures = check_file_contents(os.path.join(directory, SCRIPT_NAME), EXPECTED_LUNAR_PYTHON_SCRIPT.format(**{'reqs_to_get': self.reqs_to_get, 'res_files': self.res_files, 'output_file': self.output_file, 't_min': self.t_min, 't_max': self.t_max}))
        self.assertListEqual(failures, [])

    def test_lunar_y_file(self):
        """Tests that the lunar Y file is written correctly.

        """
        self.res_output_files = get_lunar_results(self.res_files, self.reqs_to_get, self.t_min, self.t_max, self.output_file)

        failures = []

        for res_output_file, expected_contents in zip(self.res_output_files, EXPECTED_LUNAR_Y_FILES):
            failures.append(check_file_contents(res_output_file, expected_contents))

        self.assertListEqual(failures, [])

    def tearDown(self):
        for file in self.res_output_files:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass
            except PermissionError:
                pass

EXPECTED_LUNAR_Y_FILES = [
    '''
''',
    '''
'''
]

#pylint: disable=anomalous-backslash-in-string
EXPECTED_LUNAR_PYTHON_SCRIPT = '''import os

import numpy as np
import Adams

def get_lunar_results(res_files, reqs_to_get, t_min, t_max, output_file):

    res_dict = {{}}

    for i_res, res_file in enumerate(res_files):
        
        Adams.execute_cmd('def com echo=on')

        # Load Results
        Adams.execute_cmd(f'file results read file_name="{{res_file}}"')

        # Get the model object
        mod = Adams.Models.get([mn for mn in Adams.Models][0])
        
        # Get the analysis object
        ans = mod.Analyses.get([an for an in mod.Analyses][0])
            
        # Get time and set indices of results to export 
        time = ans.results.get('TIME').values
        _time_np = np.asarray(time)
        i_min = np.argmax(_time_np >= t_min)
        i_max = np.argmax(_time_np >= t_max)

        if i_res == 0:
            # Get a name for the output time file
            time_output_file = os.path.splitext(output_file)[0] + '_time' + os.path.splitext(output_file)[-1]

            # Write the time values
            with open(time_output_file, 'w') as fid:
                fid.write(f'{{time}}'.replace('[','').replace(']','').replace(' ',''))
        
        # Define first two columns of the output string
        res_dict[os.path.split(res_file)[-1]] = {{}}
        
        for res_name, res_comps in reqs_to_get.items():

            # Get the result object handle
            res = ans.results.get(res_name)

            for res_comp in res_comps:

                values = res.get(res_comp).values[i_min:i_max]

                full_res_name = f'{{res_name}}_{{res_comp}}'

                # Get a name for the output time file
                res_output_file = os.path.splitext(output_file)[0] + '_' + full_res_name + os.path.splitext(output_file)[-1]

                with open(res_output_file, 'w' if i_res==0 else 'a') as fid:
                    fid.write(f'{{values}}\\n'.replace('[','').replace(']','').replace(' ',''))    

        # Delete the model and results
        mod.destroy()          
                
if __name__ == 'aview_main':
    
    Adams.execute_cmd('def com echo=on')

    res_files = {res_files}
    reqs_to_get = {reqs_to_get}
    t_min = {t_min}
    t_max = {t_max}
    output_suffix = r'{output_file}'

    get_lunar_results(res_files, reqs_to_get, t_min, t_max, output_suffix)
    
    print('MESSAGE:  COMPLETE')
    sys.stdout.flush()
    Adams.execute_cmd('quit conf=no')

'''


import unittest

from adamspy.adripy.utilities import get_motor_curve

import matplotlib.pyplot as plt

class Test_MotorCurveGenerator(unittest.TestCase):

    def setUp(self):
        return

    def test_motor_curve(self):
        # Set the parameters
        stall_torque = 20
        max_torque = 15
        max_rpm = 100
        min_rpm = 80

        # Get the curves
        nor_rpm, esr_rpm, nor_torque, esr_torque = motor_curve(stall_torque, max_torque, max_rpm, min_rpm)
        
        # Plot the curves
        plt.plot([rpm + min_rpm for rpm in nor_rpm], nor_torque)
        plt.plot(esr_rpm, esr_torque, linestyle='--')        

        expected_nor_torque = [
            15.0, 
            14.9625, 
            14.85, 
            14.6625, 
            14.4, 
            14.0625, 
            13.65, 
            13.1625, 
            12.6, 
            11.9625, 
            11.25, 
            10.4625, 
            9.600000000000001, 
            8.662500000000001, 
            7.65, 
            6.5625, 
            5.4, 
            4.1625, 
            2.8499999999999996, 
            1.4625000000000004, 
            0.0]

        self.assertListEqual(nor_torque, expected_nor_torque)

    def tearDown(self):
        return


"""This module is for sending scripts through Adams/PPT

"""
import os
import subprocess
import re
import time

import jinja2

LOG_COMPLETE_PATTERN = '! Command file is exhausted, batch run is finished.'

SCRIPT_NAME = '_get_lunar_results.py'
LOG_NAME = 'aview.log'

TMPLT_ENV = jinja2.Environment(
    loader=jinja2.PackageLoader('adamspy.postprocess', 'aview_scripts'),
    autoescape=False,
    keep_trailing_newline=True,
    trim_blocks=True,
    lstrip_blocks=True
)

def get_lunar_results(res_files, reqs_to_get, t_min, t_max, output_file, _just_write_script=False):
    
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', SCRIPT_NAME)).read())
    
    working_directory = os.path.dirname(res_files[0])

    with open(os.path.join(working_directory, SCRIPT_NAME), 'w') as fid:
        fid.write(template.render({'res_files': res_files, 'reqs_to_get': reqs_to_get, 't_min': t_min, 't_max': t_max, 'output_suffix': output_file}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, 'aview.log'))
        except FileNotFoundError:
            pass

        # Run the postprocessor
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW        
        subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], SCRIPT_NAME), cwd=working_directory, startupinfo=startupinfo)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME))

        os.remove(os.path.join(working_directory, SCRIPT_NAME))

        # Make a list of the files that are written
        res_output_files = [os.path.splitext(output_file)[0] + '_time' + os.path.splitext(output_file)[-1]]
        for res_name, res_comps in reqs_to_get.items():
            for res_comp in res_comps:
                full_res_name = f'{res_name}_{res_comp}'
                res_output_files.append(os.path.splitext(output_file)[0] + '_' + full_res_name + os.path.splitext(output_file)[-1])

        return res_output_files

def _wait(log_file, sleep_time=0.2, timeout=60):
    """Waits for the log file to write the last line of the macro
    
    Parameters
    ----------
    log_file : str
        filename of log file
    sleep_time : float, optional
        Time between checks, by default 0.2
    timeout : int, optional
        During after which to time out, by default 60

    """    

    for _i in range(int(timeout/sleep_time)):
        
        ppt_log_file_exists = os.path.exists(log_file)       

        if ppt_log_file_exists is True:
            # If ppt.log exists, open it and see if the results have been loaded
            with open(log_file, 'r') as fid:
                text = fid.read()
            if re.search(LOG_COMPLETE_PATTERN, text):
                break
                
        time.sleep(sleep_time)



"""Code for generating animations automatically in Adams View

"""
import os
import re
import time
import subprocess

from . import AVIEW_AFTERSTART_FILENAME

LOG_COMPLETE_PATTERN = ' >> quit conf=no'

def create_animation(model_file : str, res_file : str, model_name : str, analysis_name : str, timeout=60):
    """Creates an animation from the given model.
    
    Parameters
    ----------
    model_file : str
        Adams View model (.cmd) file
    res_file : str
        Adams results (.res) file
    model_name : str
        Name of the model
    analysis_name : str
        Name of the analysis
    wait : bool, optional
        If `True`, code execution will freeze until the animation script completes and the postprocessor closes. (the default is True)    

    """
    # Determine the directory, relative results filename, and pptAS name
    directory, res_file = os.path.split(res_file)
    if not directory:
        directory = os.getcwd()    

    ppt_as_filename = os.path.join(directory, AVIEW_AFTERSTART_FILENAME)

    # Write aviewAS.cmd
    with open(ppt_as_filename, 'w') as fid:
        fid.write(CMD_CODE.format(**{'model_file': model_file, 'res_file': res_file, 'model_name': model_name, 'analysis_name': analysis_name, 'mov_filename': analysis_name}))

    # Set the ppt.log filename
    ppt_log_file = os.path.join(directory, 'aview.log')

    # Remove aview.log
    if os.path.exists(ppt_log_file):
        try:
            os.remove(ppt_log_file)
        except PermissionError:
            pass

    # Run the postprocessor
    startupinfo = subprocess.STARTUPINFO()
    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW        
    _ppt_proc = subprocess.Popen('"{}" aview ru-s i'.format(os.environ['ADAMS_LAUNCH_COMMAND']), cwd=directory, startupinfo=startupinfo)
    
    # Wait for process to complete
    _wait(ppt_log_file, timeout=timeout)

    # Remove the pptAS file
    try:
        os.remove(ppt_as_filename)
    except PermissionError:
        pass

    return os.path.join(directory, analysis_name + '.avi')

def _wait(log_file, sleep_time=0.2, timeout=60):
    """Waits for the log file to write the last line of the macro
    
    Parameters
    ----------
    log_file : str
        filename of log file
    sleep_time : float, optional
        Time between checks, by default 0.2
    timeout : int, optional
        During after which to time out, by default 60

    """    
    for _i in range(int(timeout/sleep_time)):
        
        ppt_log_file_exists = os.path.exists(log_file)       

        if ppt_log_file_exists is True:
            # If ppt.log exists, open it and see if the results have been loaded
            with open(log_file, 'r') as fid:
                text = fid.read()
            if re.search(LOG_COMPLETE_PATTERN, text):
                break
                
        time.sleep(sleep_time)

CMD_CODE = '''
!
def com echo=on
file command read file_name = "{model_file}"
file results read model_name = {model_name} file_name = "{res_file}" 
interface plot window open
interface window undisplay window = .gui.main
!
!
interface plot window page_display page=.gui.ppt_main.sash1.sash2.gfx.page_1
!****** These commands create the animation movie file.
! !
 if cond=("animation" !="")
   interface plot window set_mode mode=Animation
 else
   interface plot window set_mode=plotting
 end

 interface plot window load_view analysis=.{model_name}.{analysis_name}

 color modify &
   color=Background &
   red=1.0 &
   green=1.0 &
   blue=1.0
 !
 view manage orient view=iso up_axis=(.MDI.up_axis) forward_axis=(.MDI.forward_axis)
 view zoom auto=on
!
 animation modify increment=10 animation_name=.animation_1
 animation record start image_type = avi avi_compression = yes avi_frames_per_sec = 10 image_prefix = "{mov_filename}.avi" avi_keyframe_every = 500 avi_quality = 75
 defaults animation_settings loop=play_once
 animation play page_name=.gui.ppt_main.sash1.sash2.gfx.page_1
 animation record stop 
!

quit conf=no
'''
